#

/*
 * DIVA disk driver, written by R. Halstead, MIT, June 1976.
 *
 * Re-written by N. Chiappa, MIT-LCS-CSR  12-4-79.
 *
 */

#include "../h/param.h"
#include "../h/systm.h"
#include "../h/buf.h"
#include "../h/conf.h"
#include "../h/user.h"
#include "../h/userx.h"


/* DIVA disk regs */

struct {
	int	dvcsr;
	int	dvdbr;
	int	dvmar;
	int	dvcwcr;
	int	dvcbr;
	int	dvssr;
	int	dvair;
	int	dvusr;
       };


/* Controller bits - CSR first, then SSR, then CBR commands */

#define	DVGO	01			/* go */
#define	DVIE	0100			/* enable interrupt */
#define	DVIEGO	(DVIE | DVGO)		/* go, enable interrupt */
#define	DVDONE	0200			/* done (for clearing) */
#define	DVDTLT	01000			/* data late err bit */
#define	DVSLCK	02000			/* select lock */
#define	DVCNBSY	04000			/* controller busy */
#define	DVCHE	(DVSLCK | 070000)	/* hard err bits in csr */
#define	DVERR	0100000			/* error bit in csr */

#define	DVPROT	02000			/* disk is wrt protected */
#define	DVRDY	04000			/* drive ready in */
#define	DVONL	010000			/* drive online */
#define	DVLRG	020000			/* double density drive */
#define	DVSHE	0140000			/* hard err bits in ssr */


#define	DVDSEL	0140000			/* select drive */
#define	DVLDCY	020000			/* select cyl */
#define	DVSKRH	030060			/* seek and reset head */
#define	DVNXSRF	030001			/* next surface */
#define	DVRD	070000			/* cnfrn and read */
#define	DVWRT	0130000			/* cnfrn and write */
#define	DVCDA	030100			/* clr drv attn */
#define	DVRCAL	030002			/* recalibrate */
#define	DVRSET	0170000			/* reset */


/* Useful functions - middle group for hacking confirm hdrs */

#define	dvdev(i)	(i & 037)		/* mask off drive number */
#define	dvdrv(i)	((i >> 5) & 07)		/* get driver number */

#define	cyl(i)		(i & 0777)		/* get cyl from header word */
#define	putcyl(i)	(i)			/* for clarity */
#define	surf(i)		(i >> 9)		/* get surf */
#define	putsurf(i)	(i << 9)		/* put surf in hdr */
#define	sect(i)		(i & 017)		/* get sect */
#define putsect(i)	(i)			/* clarity too */

#define	wtob(i)		(i >> 8)		/* words to blocks */
#define	btob(i)		(i >> 9)		/* bytes to blocks */
#define	dvwait()	while ((DVADDR->dvcsr & DVCNBSY) != 0)


/* useful constants */

#define DVMXHEAD	20			/* # of heads on DIVA drive */
#define DVMXSECT	12			/* # of sectors per track */
#define	BLKSIZ		256			/* 256 words in blk */
#define	HBLKSIZ		(-BLKSIZ << 4)		/* for header */
#define	NEWCYL		-1			/* command for new cyl */
#define	DVSKI		0			/* seek int */
#define	DVRWI		1			/* read/write int */
#define	DVMXDRV		4			/* # of drives handled (max) */


/* Tuneable params */

#define DVMXERR	5			/* max retries on error */
#define	DK_N	3			/* monitoring device bit */
#define DVADDR	0164000			/* addr of dvcsr */
#define NDV	2			/* number of drives */
#define	NDEV	25			/* number of devices per drive.. */


struct	devtab	dvtab;
struct	buf	rdvbuf;			/* for raw I/O */

/* dvstrat uses b_error to save sector number,
 * b_resid to save cyl + head number.
 */

#define sectno	b_error
#define cylsurf	b_resid


struct {				/* static for disk handler */
	int	dvcylsurf;		/* cyl & srf #, 1st wrd of cnfrm buf */
	int	dvsectno;		/* sect # & word cnt, 2nd wrd of buf */
	int	dvchecksum;		/* confirm buffer checksum */
	char	*dvmemaddr;		/* phys buffer addr for r/w data */
	char	*wcount;		/* negative # of words yet to xfer */
	int	inthand;		/* 0 = dvrw, 1 = dvrwx : intr hndlr */
	int	prvdrv;			/* previous drive */
	} dvdata	{ 0, 0, 0, 0, 0, 0, -1 };


struct	{				/* table of disk configurations */
	char	*nblocks;
	int	cyloff;
	}	dv_sizes[]	{	/*		     minor #	*/

	48720,	0,	/* cyl 0 thru 202			0	*/
	48720,	203,	/* cyl 203 thru 405			1	*/
	31680,	0,	/* cyl 0 thru 131	dv1, dv11	2	*/
	28800,	0,	/* cyl 360 thru 405			3	*/
	65535,	0,	/* cyl 0 thru 273 			4	*/
	65535,	132,	/* cyl 132 thru 405	dv0, dv10	5	*/
	4800,	0,	/* meta-disk 0				6	*/
	4800,	20,	/* meta-disk 1				7	*/
	4800,	40,	/* meta-disk 2				8	*/
	4800,	60,	/* 3					9	*/
	4800,	80,	/* 4					10	*/
	4800,	100,	/* 5					11	*/
	4800,	120,	/* 6					12	*/
	4800,	140,	/* 7					13	*/
	4800,	160,	/* 8					14	*/
	4800,	180,	/* 9					15	*/
	4800,	200,	/* 10					16	*/
	4800,	220,	/* 11					17	*/
	4800,	240,	/* 12					18	*/
	4800,	260,	/* 13					19	*/
	4800,	280,	/* 14					20	*/
	4800,	300,	/* 15					21	*/
	4800,	320,	/* 16					22	*/
	4800,	340,	/* 17					23	*/
	4800,	360,	/* 18					24	*/
	4800,	380,	/* 19					25	*/
	};


/* Main DIVA strategy routine. Check to see that he has a legal drive,
 * meta-disk and blk, then convert his block number to an abs block number.
 * Convert this to a cyl/surf/sect coordinate, store that info away
 * in the header, sort into the request list (use elevator algorithm
 * at present) and start disk if idle.
 */

dvstrategy(abp)
struct buf *abp;

{	register struct buf *bp;
	register char *p1, *p2;
	int	drive;

	bp = abp;

	drive = dvdrv(bp->b_dev.d_minor);
	if (drive > DVMXDRV) goto lose;
	if (drive > NDV) goto lose;
	if (dvdev(bp->b_dev.d_minor) > NDEV) goto lose;
	p1 = &dv_sizes[dvdev(bp->b_dev.d_minor)];

	if ((bp->b_blkno >= p1->nblocks) ||
	    ((bp->b_blkno + wtob(-bp->b_wcount)) > p1->nblocks)) {
lose:		bp->b_flags =| B_ERROR;
		bp->b_error = ENXIO;
		iodone(bp);
		return;
		}

	bp->av_forw = NULL;

	bp->cylsurf = putcyl(p1->cyloff);	/* cylinder dev starts */
	p1 = bp->b_blkno;
	p2 = lrem(p1, DVMXSECT*DVMXHEAD);	/* sect and surface */
	p1 = ldiv(p1, DVMXSECT*DVMXHEAD);	/* cylinder within dev */
	bp->cylsurf =+ putcyl(p1);		/* get real cylinder */
	p1 = ldiv(p2, DVMXSECT);		/* get surface */
	bp->cylsurf =| putsurf(p1);		/* and sector */
	bp->sectno = putsect(lrem(p2, DVMXSECT));

	spl5();
	if ((p1 = dvtab.d_actf) == NULL) dvtab.d_actf = bp;
	 else {
		for (; p2 = p1->av_forw ; p1 = p2) {
			if (((cyl(p1->cylsurf) <= cyl(bp->cylsurf)) &&
			     (cyl(bp->cylsurf) < cyl(p2->cylsurf))) ||
			    ((cyl(p1->cylsurf) >= cyl(bp->cylsurf)) &&
			     (cyl(bp->cylsurf) > cyl(p2->cylsurf)))) break;
			}
		bp->av_forw = p2;
		p1->av_forw = bp;
	      }
	if (dvtab.d_active == 0) dvstart();
	spl0();
}


/* Routine to start disk operation. Select drive if necessary, then check to
 * make sure drive is operational and not write prot on a write request, as
 * operating in those circumstances hangs the drive. Start the seek and
 * return - the I/O will be done when the seek interrupt arrives.
 * Note: only allocate map if we need it and don't already have it; this
 * routine can be called from the error handler with the map already
 * allocated.
 */

dvstart()

{	register	struct buf	*bp;
	register	tmp;

	if ((bp = dvtab.d_actf) == NULL)
		return;
	dvtab.d_active++;

	if (((bp->b_flags & B_PHYS) != 0) && ((bp->b_flags & B_MAP) == 0))
		mapalloc(bp);

	if ((tmp = dvdrv(bp->b_dev.d_minor)) != dvdata.prvdrv ) {
		dvdata.prvdrv = tmp;
		DVADDR->dvcbr = (DVDSEL | tmp);
		dvwait();
		}

	tmp = ~DVADDR->dvssr;
	if ((DVADDR->dvcsr & DVSLCK) || ((tmp & DVONL) == 0) ||
	    ((tmp & DVPROT) && !(bp->b_flags & B_READ))) {
		deverror(bp, DVADDR->dvcsr, tmp);
		bp->b_flags =| B_ERROR;
		if ((tmp & DVPROT) && !(bp->b_flags & B_READ))
			bp->b_error = EROFS;
		  else
			bp->b_error = EIO;		
		dvend();
		return;
		}

	dvseek1();
	dk_busy =| 1<<DK_N;
	dk_numb[DK_N] =+ 1;
	tmp = ((unsigned)(-bp->b_wcount) >> 5);
	dk_wds[DK_N] =+ tmp;
}


/* Set up confirm buffer using data in buffer header and start seek.
 */

dvseek1()

{	register	struct buf	*bp;

	bp = dvtab.d_actf;
	dvdata.inthand = DVSKI;
	dvdata.dvcylsurf = bp->cylsurf;
	dvdata.dvsectno = bp->sectno;
	dvdata.dvsectno =| HBLKSIZ;		/* always 1 block at a time */
	dvdata.wcount = bp->b_wcount;
	dvdata.dvmemaddr = bp->b_addr;
	DVADDR->dvcsr = (bp->b_xmem << 4);
	dvseek2();
}


/* Seek to sector in confirm buffer. Called at start of transfer from
 * dvseek1 and during transfer from dvrwx if a transfer overflows
 * from one cylinder to the next.
 */

dvseek2()

{	register int i;

	DVADDR->dvcbr = cyl(dvdata.dvcylsurf) | DVLDCY;
	dvwait();
	DVADDR->dvcbr = DVCDA;
	dvwait();
	DVADDR->dvair = 1 << dvdrv(dvtab.d_actf->b_dev.d_minor);
	DVADDR->dvcsr =| DVDONE;		/* clear poss spurious int */
	DVADDR->dvcsr =| DVIE;
	DVADDR->dvcbr = DVSKRH;
}


/* Seek complete interrupt handler - do final controller setup
 * and the start I/O operation.
 */

dvrw()

{	register	sct;

	sct = dvrw1();
	dvdata.inthand = DVRWI;
	dvrwv(sct);
}


/* Initiate transfer of data. This called by dvrw after a seek completes
 * and by dvrwx when in the middle of multi-block transfers.
 */

dvrwv(sct)

{	register	cmd;

	cmd = sct << 1;		/* Two phys sects per virtual */
	cmd =| ((dvtab.d_actf->b_flags & B_READ) ? DVRD : DVWRT);
	DVADDR->dvcsr =| DVDONE;
	DVADDR->dvcbr = cmd;
	DVADDR->dvcsr =| DVIEGO;
}


/* Handle xfer complete interrupts - if we have transferred all the data,
 * end things off, else get a new command from dvrw2 (which will alter
 * the confirm buffer as necessary). If we are still on the same cylinder,
 * go do the next block, otherwise call for a seek.
 */

dvrwx()

{	register	sct;

	if (dvdata.wcount == 0) {
		dvdata.wcount = BLKSIZ;
		dvend();
		return;
		}

	sct = dvrw2();
	if (sct != NEWCYL) {
		dvrwv(sct);
		return;
		}

	dvdata.inthand = DVSKI;
	dvseek2();
}


/* Set for xfer right after seek and then return current sector number.
 */

dvrw1()

{	DVADDR->dvcbr = surf(dvdata.dvcylsurf);
	dvwait();
	return(dvrw3());
}


/* Set up disk regs for next xfer, do transfer length accounting,
 * set up current tranfer length and return sector number for use
 * by dvrwv.
 */

dvrw3()

{	dvdata.dvchecksum = -dvdata.dvcylsurf - dvdata.dvsectno;
	DVADDR->dvmar = &dvdata;
	if (dvdata.wcount > -BLKSIZ) {
		DVADDR->dvcwcr = dvdata.wcount << 1;
		dvdata.wcount = 0;
		}
	 else {
		DVADDR->dvcwcr = -(BLKSIZ << 1);	/* byte cnt reg */
		dvdata.wcount =+ BLKSIZ;
		}
	return(sect(dvdata.dvsectno));
}


/* Set up confirm header to resume xfer at next sector in cases where all
 * the data has not been transferred. Check for sector and surface overflows,
 * and in the latter case call for a seek, and in the former go to the next
 * surface. Returns sector number or NEWCYL if new seek needed.
 */

dvrw2()

{	register	i;

	dvdata.dvmemaddr = DVADDR->dvmar;

	i = sect(dvdata.dvsectno);
	dvdata.dvsectno = HBLKSIZ;
	if (++i < DVMXSECT) {
		dvdata.dvsectno =| putsect(i);
		return(dvrw3());
		}

	i = surf(dvdata.dvcylsurf);
	dvdata.dvcylsurf = putcyl(cyl(dvdata.dvcylsurf));
	if (++i >= DVMXHEAD) {
		dvdata.dvcylsurf =+ putcyl(1);
		return(NEWCYL);
		}

	DVADDR->dvcbr = DVNXSRF;
	dvdata.dvcylsurf =| putsurf(i);
	dvwait();
	return(dvrw3());
}


/* DIVA disk interrupt handler - check to make sure that there is something
 * going on, then continue. If there was an error on the transfer, log it.
 * if the error was a soft error, then retry the transfer a few times.
 * If the error was possibly due to a bad seek, recalibrate the drive.
 * (Note that due to the losing controller we can't differentiate between
 * header mismatches and checksum errors, so in both cases we wind up
 * doing a recalibrate.) If the error was a hard error, reset the controller
 * and recalibrate the drive. If there was no error, simply hand off to the
 * right handler.
 */

dvintr()

{	register	struct buf	*bp;
	register	ctr, csr;

	if (dvtab.d_active == 0) {
		printf("\nrandom DIVA int\n");
		return;
		}

	bp = dvtab.d_actf;

	if (DVADDR->dvcsr & DVERR) {
		dk_errs[DK_N] =+ 1;
		ctr = ~DVADDR->dvssr;
		csr = DVADDR->dvcsr;
		deverror(bp, csr, ctr);

	   	if ((ctr & DVSHE) || (csr & DVCHE)) {
			DVADDR->dvcbr = DVRSET;
			dvwait();
			if (!(csr & DVSLCK))
				dvcal();
			bp->b_flags =| B_ERROR;
			bp->b_error = EIO;
			dvend();
			return;
			}

		if ((csr & DVDTLT) == 0) 	/* don't recal if slow NPR */
			dvcal();		/* would be nice if no CRC */

		if (++dvtab.d_errcnt < DVMXERR) {
			dk_busy =& ~(1<<DK_N);
			dvstart();
			return;
			}

		bp->b_flags =| B_ERROR;
		bp->b_error = EIO;
		dvend();
		return;
		}

	if (dvdata.inthand == DVRWI)
		dvrwx();
	 else dvrw();
}

/* Recalibrate drive after error.
 */

dvcal()

{		DVADDR->dvcbr = DVRCAL;
		dvwait();
		while (DVADDR->dvssr & DVRDY);
}


/* Clean up after xfer; signal I/O complete and restart
 * next I/O.
 */

dvend()

{	register	struct buf	*bp;

	dk_busy =& ~(1<<DK_N);
	bp = dvtab.d_actf;
	dvtab.d_errcnt = 0;
	dvtab.d_active = 0;
	dvtab.d_actf = bp->av_forw;
	bp->b_resid = dvdata.wcount - BLKSIZ;
	iodone(bp);
	dvstart();
}


/* Raw DIVA read and write drivers.
 */

dvread(dev)
{	if (dvphys(dev))
		physio(&dvstrategy, &rdvbuf, dev, B_READ);
}

dvwrite(dev)
{	if (dvphys(dev))
		physio(&dvstrategy, &rdvbuf, dev, B_WRITE);
}


/* Check DIVA disk address for validity; used by raw DIVA
 * access routines.
 */

dvphys(dev)

{	register	blk, drive;

	drive = dvdrv(dev.d_minor);
	if (drive > DVMXDRV) goto lose;
	if (drive > NDV) goto lose;
	if (dvdev(dev.d_minor) > NDEV) goto lose;

	blk = lshift(u.u_offset, -9);
	blk =+ ldiv((u.u_count + (BLKSIZ*2 -1)), BLKSIZ*2);
	if (blk > dv_sizes[dvdev(dev.d_minor)].nblocks) goto lose;
	return(1);

lose:	u.u_error = ENXIO;
	return(0);
}
